@@ -25,20 +25,29 @@ class Clustergram:
     as k-means and for hierarchical cluster algorithms when the number of
     observations is large enough to make dendrograms impractical.
 
-    Clustergram offers two backends for the computation - ``scikit-learn``
-    which uses CPU and RAPIDS.AI ``cuML``, which uses GPU. Note that both
-    are optional dependencies, but you will need at least one of them to
+    Clustergram offers three backends for the computation - ``scikit-learn`` and
+    ``scipy`` which use CPU and RAPIDS.AI ``cuML``, which uses GPU. Note that all
+    are optional dependencies but you will need at least one of them to
     generate clustergram.
 
+    Alternatively, you can create clustergram using ``from_data`` or
+    ``from_centers`` methods based on alternative clustering algorithms.
+
     Parameters
     ----------
-    k_range : iterable
+    k_range : iterable (default None)
         iterable of integer values to be tested as ``k`` (number of cluster or
-        components).
-    backend : {'sklearn', 'cuML'} (default 'sklearn')
-        Whether to use ``sklearn``'s implementation of KMeans and PCA or ``cuML``
-        version. ``sklearn`` does computation on CPU, ``cuml`` on GPU.
-    method : {'kmeans', 'gmm', 'minibatchkmeans'} (default 'kmeans')
+        components). Not required for hierarchical clustering but will be applied
+        if given. It is recommended to always use limited range for hierarchical
+        methods as unlimited clustergram can take a while to compute and for large
+        number of observations is not legible.
+    backend : {'sklearn', 'cuML', 'scipy'} (default None)
+        Specify computational backend. Defaults to ``sklearn`` for ``'kmeans'``,
+        ``'gmm'``, and ``'minibatchkmeans'`` methods and to ``'scipy'`` for any of
+        hierarchical clustering methods. ``'scipy'`` uses ``sklearn`` for PCA
+        computation if that is required.
+        ``sklearn`` does computation on CPU, ``cuml`` on GPU.
+    method : {'kmeans', 'gmm', 'minibatchkmeans', 'hierarchical'} (default 'kmeans')
         Clustering method.
 
         * ``kmeans`` uses K-Means clustering, either as ``sklearn.cluster.KMeans``
@@ -46,6 +55,8 @@ class Clustergram:
         * ``gmm`` uses Gaussian Mixture Model as ``sklearn.mixture.GaussianMixture``
         * ``minibatchkmeans`` uses Mini Batch K-Means as
           ``sklearn.cluster.MiniBatchKMeans``
+        * ``hierarchical`` uses hierarchical/agglomerative clustering as
+          ``scipy.cluster.hierarchy.linkage``. See
 
         Note that ``gmm`` and ``minibatchkmeans`` are currently supported only
         with ``sklearn`` backend.
@@ -53,7 +64,9 @@ class Clustergram:
         Print progress and time of individual steps.
     **kwargs
         Additional arguments passed to the model (e.g. ``KMeans``),
-        e.g. ``random_state``.
+        e.g. ``random_state``. Pass ``linkage`` to specify linkage method in
+        case of hierarchical clustering (e.g. ``linkage='ward'``). See the
+        documentation of scipy for details.
 
 
     Attributes
@@ -63,6 +76,8 @@ class Clustergram:
         DataFrame with cluster labels for each iteration.
     cluster_centers : dict
         Dictionary with cluster centers for each iteration.
+    linkage : scipy.cluster.hierarchy.linkage
+        Linkage object for hierarchical methods.
 
 
     Examples
@@ -91,7 +106,12 @@ class Clustergram:
     """
 
     def __init__(
-        self, k_range, backend="sklearn", method="kmeans", verbose=True, **kwargs,
+        self,
+        k_range=None,
+        backend=None,
+        method="kmeans",
+        verbose=True,
+        **kwargs,
     ):
         self.k_range = k_range
 
@@ -99,15 +119,21 @@ class Clustergram:
         kwargs.pop("pca_weighted", None)
         kwargs.pop("pca_kwargs", None)
 
-        if backend not in ["sklearn", "cuML"]:
+        if backend is None:
+            backend = "scipy" if method == "hierarchical" else "sklearn"
+
+        allowed_backends = ["sklearn", "cuML", "scipy"]
+        if backend not in allowed_backends:
             raise ValueError(
-                f'"{backend}" is not a supported backend. Use "sklearn" or "cuML".'
+                f'"{backend}" is not a supported backend. '
+                f"Use one of {allowed_backends}."
             )
         else:
             self.backend = backend
 
-        supported = ["kmeans", "gmm", "minibatchkmeans"]
-        if method not in supported:
+        supported = ["kmeans", "gmm", "minibatchkmeans", "hierarchical"]
+        class_methods = ["from_centers", "from_data"]
+        if method not in supported and method not in class_methods:
             raise ValueError(
                 f"'{method}' is not a supported method. "
                 f"Only {supported} are supported now."
@@ -115,10 +141,23 @@ class Clustergram:
         else:
             self.method = method
 
+        if self.k_range is None and self.method != "hierarchical":
+            raise ValueError(f"'k_range' is mandatory for '{self.method}' method.")
+
+        if (
+            (self.backend == "cuML" and self.method != "kmeans")
+            or (self.backend == "scipy" and self.method != "hierarchical")
+            or (self.backend == "sklearn" and self.method == "hierarchical")
+        ):
+            raise ValueError(
+                f"'{self.method}' method is not implemented "
+                f"for '{self.backend}' backend. Use supported combination."
+            )
+
         self.engine_kwargs = kwargs
         self.verbose = verbose
 
-        if self.backend == "sklearn":
+        if self.backend in ["sklearn", "scipy"]:
             self.plot_data_pca = pd.DataFrame()
             self.plot_data = pd.DataFrame()
         else:
@@ -164,9 +203,11 @@ class Clustergram:
             elif self.method == "minibatchkmeans":
                 self._kmeans_sklearn(data, minibatch=True, **kwargs)
             elif self.method == "gmm":
-                self.means = self._gmm_sklearn(data, **kwargs)
+                self._gmm_sklearn(data, **kwargs)
         if self.backend == "cuML":
-            self.means = self._kmeans_cuml(data, **kwargs)
+            self._kmeans_cuml(data, **kwargs)
+        if self.backend == "scipy":
+            self._scipy_hierarchical(data, **kwargs)
 
     def _kmeans_sklearn(self, data, minibatch, **kwargs):
         """Use scikit-learn KMeans"""
@@ -217,7 +258,6 @@ class Clustergram:
         """Use sklearn.mixture.GaussianMixture"""
         try:
             from sklearn.mixture import GaussianMixture
-            import numpy as np
             from scipy.stats import multivariate_normal
         except ImportError:
             raise ImportError(
@@ -250,6 +290,170 @@ class Clustergram:
 
             print(f"K={n} fitted in {time() - s} seconds.") if self.verbose else None
 
+    def _scipy_hierarchical(self, data, **kwargs):
+        """Use scipy.cluster.hierarchy.linkage"""
+        try:
+            from scipy.cluster import hierarchy
+        except ImportError:
+            raise ImportError("scipy is required to use `scipy` backend.")
+
+        method = self.engine_kwargs.pop("linkage", "single")
+        self.linkage = hierarchy.linkage(data, method=method, **self.engine_kwargs)
+        rootnode, nodelist = hierarchy.to_tree(self.linkage, rd=True)
+        distances = [node.dist for node in nodelist if node.dist > 0][::-1]
+
+        self.labels = pd.DataFrame()
+        self.cluster_centers = {}
+
+        if self.k_range is None:
+            self.k_range = range(1, len(distances) + 1)
+
+        if not isinstance(data, pd.DataFrame):
+            data = pd.DataFrame(data)
+
+        for i in self.k_range:
+            d = distances[i - 1]
+            lab = hierarchy.fcluster(self.linkage, d, criterion="distance")
+            self.labels[i] = lab - 1
+            self.cluster_centers[i] = data.groupby(lab).mean().values
+
+    @classmethod
+    def from_centers(cls, cluster_centers, labels, data=None):
+        """Create clustergram based on cluster centers dictionary and labels DataFrame
+
+        Parameters
+        ----------
+
+        cluster_centers : dict
+            dictionary of cluster centers with keys encoding the number of clusters
+            and values being ``M``x````N`` arrays where ``M`` == key and ``N`` ==
+            number of variables in the original dataset.
+            Entries should be ordered based on keys.
+        labels : pandas.DataFrame
+            DataFrame with columns representing cluster labels and rows representing
+            observations. Columns must be equal to ``cluster_centers`` keys.
+        data : array-like (optional)
+            array used as an input of the clustering algorithm with ``N`` columns.
+            Required for `plot(pca_weighted=True)` plotting option. Otherwise only
+            `plot(pca_weighted=False)` is available.
+
+        Returns
+        -------
+        clustegram.Clustergram
+
+        Notes
+        -----
+        The algortihm uses ``sklearn`` and ``pandas`` to generate clustergram.
+        GPU option is not implemented.
+
+        Examples
+        --------
+        >>> import pandas as pd
+        >>> import numpy as np
+        >>> labels = pd.DataFrame({1: [0, 0, 0], 2: [0, 0, 1], 3: [0, 2, 1]})
+        >>> labels
+           1  2  3
+        0  0  0  0
+        1  0  0  2
+        2  0  1  1
+        >>> centers = {
+        ...             1: np.array([[0, 0]]),
+        ...             2: np.array([[-1, -1], [1, 1]]),
+        ...             3: np.array([[-1, -1], [1, 1], [0, 0]]),
+        ...         }
+        >>> cgram = Clustergram.from_centers(centers, labels)
+        >>> cgram.plot(pca_weighted=False)
+
+        >>> data = np.array([[-1, -1], [1, 1], [0, 0]])
+        >>> cgram = Clustergram.from_centers(centers, labels, data=data)
+        >>> cgram.plot()
+
+        """
+        if not (list(cluster_centers.keys()) == labels.columns).all():
+            raise ValueError("'cluster_centers' keys do not match 'labels' columns.")
+
+        cgram = cls(k_range=list(cluster_centers.keys()), method="from_centers")
+
+        cgram.cluster_centers = cluster_centers
+        cgram.labels = labels
+        cgram.backend = "sklearn"
+
+        if data is not None:
+            cgram.data = data
+
+        return cgram
+
+    @classmethod
+    def from_data(cls, data, labels, method="mean"):
+        """Create clustergram based on data and labels DataFrame
+
+        Cluster centers are created as mean values or median values as a
+        groupby function over data using individual labels.
+
+        Parameters
+        ----------
+
+        data : array-like
+            array used as an input of the clustering algorithm in the ``(M, N)`` shape
+            where ``M`` == number of observations and ``N`` == number of variables
+        labels : pandas.DataFrame
+            DataFrame with columns representing cluster labels and rows representing
+            observations. Columns must be equal to ``cluster_centers`` keys.
+        method : {'mean', 'median'}, default 'mean'
+            Method of computation of cluster centres.
+
+        Returns
+        -------
+        clustegram.Clustergram
+
+        Notes
+        -----
+        The algortihm uses ``sklearn`` and ``pandas`` to generate clustergram.
+        GPU option is not implemented.
+
+        Examples
+        --------
+        >>> import pandas as pd
+        >>> import numpy as np
+        >>> data = np.array([[-1, -1, 0, 10], [1, 1, 10, 2], [0, 0, 20, 4]])
+        >>> data
+        array([[-1, -1,  0, 10],
+               [ 1,  1, 10,  2],
+               [ 0,  0, 20,  4]])
+        >>> labels = pd.DataFrame({1: [0, 0, 0], 2: [0, 0, 1], 3: [0, 2, 1]})
+        >>> labels
+           1  2  3
+        0  0  0  0
+        1  0  0  2
+        2  0  1  1
+        >>> cgram = Clustergram.from_data(data, labels)
+        >>> cgram.plot()
+
+        """
+
+        cgram = cls(k_range=list(labels.columns), method="from_data")
+
+        cgram.cluster_centers = {}
+        cgram.data = data
+
+        if not isinstance(data, pd.DataFrame):
+            data = pd.DataFrame(data)
+
+        for i in cgram.k_range:
+            if method == "mean":
+                cgram.cluster_centers[i] = data.groupby(labels[i]).mean().values
+            elif method == "median":
+                cgram.cluster_centers[i] = data.groupby(labels[i]).median().values
+            else:
+                raise ValueError(
+                    f"'{method}' is not supported. Use 'mean' or 'median'."
+                )
+
+        cgram.labels = labels
+        cgram.backend = "sklearn"
+
+        return cgram
+
     def silhouette_score(self, **kwargs):
         """
         Compute the mean Silhouette Coefficient of all samples.
@@ -268,7 +472,7 @@ class Clustergram:
 
         Returns
         -------
-        self.silhouette : pd.Series
+        silhouette : pd.Series
 
         Notes
         -----
@@ -304,7 +508,7 @@ class Clustergram:
 
         self.silhouette = pd.Series(name="silhouette_score", dtype="float64")
 
-        if self.backend == "sklearn":
+        if self.backend in ["sklearn", "scipy"]:
             for k in self.k_range:
                 if k > 1:
                     self.silhouette.loc[k] = metrics.silhouette_score(
@@ -337,7 +541,7 @@ class Clustergram:
 
         Returns
         -------
-        self.calinski_harabasz : pd.Series
+        calinski_harabasz : pd.Series
 
         Notes
         -----
@@ -375,7 +579,7 @@ class Clustergram:
             name="calinski_harabasz_score", dtype="float64"
         )
 
-        if self.backend == "sklearn":
+        if self.backend in ["sklearn", "scipy"]:
             for k in self.k_range:
                 if k > 1:
                     self.calinski_harabasz.loc[k] = metrics.calinski_harabasz_score(
@@ -405,7 +609,7 @@ class Clustergram:
 
         Returns
         -------
-        self.davies_bouldin : pd.Series
+        davies_bouldin : pd.Series
 
         Notes
         -----
@@ -441,7 +645,7 @@ class Clustergram:
 
         self.davies_bouldin = pd.Series(name="davies_bouldin_score", dtype="float64")
 
-        if self.backend == "sklearn":
+        if self.backend in ["sklearn", "scipy"]:
             for k in self.k_range:
                 if k > 1:
                     self.davies_bouldin.loc[k] = metrics.davies_bouldin_score(
@@ -569,13 +773,13 @@ class Clustergram:
             if self.plot_data_pca.empty:
                 pca_kwargs.pop("n_components", None)
 
-                if self.backend == "sklearn":
+                if self.backend in ["sklearn", "scipy"]:
                     self._compute_pca_means_sklearn(**pca_kwargs)
                 else:
                     self._compute_pca_means_cuml(**pca_kwargs)
         else:
             if self.plot_data.empty:
-                if self.backend == "sklearn":
+                if self.backend in ["sklearn", "scipy"]:
                     self._compute_means_sklearn()
                 else:
                     self._compute_means_cuml()
@@ -612,7 +816,7 @@ class Clustergram:
         for i in k_range:
             cl = means[i].value_counts()
 
-            if self.backend == "sklearn":
+            if self.backend in ["sklearn", "scipy"]:
                 ax.scatter(
                     [i] * i,
                     [cl.index],
@@ -636,7 +840,7 @@ class Clustergram:
                 )
 
             try:
-                if self.backend == "sklearn":
+                if self.backend in ["sklearn", "scipy"]:
                     sub = means.groupby([i, i + 1]).count().reset_index()
                 else:
                     sub = means.groupby([i, i + 1]).count().reset_index().to_pandas()
